#!/usr/bin/perl -w

# .CON analysis program
# by Eric Haines, erich@acm.org
# September 27, 2003

# This Perl program analyzes all .con and .inc files found in the directory
# and subdirectories from where the code is executed. It generates three
# reports:
#
# 1) a list of all properties found and lists of values used for setting each.
# 2) for each property, a list of directories and files associated with it.
# 3) for each type of .con file, a list of the properties and their frequency of use.
#
# You'll need Perl to run the conanalysis.pl program. Just get it for free
# from Activestate, at
# http://www.activestate.com/Products/Download/Register.plex?id=ActivePerl
#
# To run, put conanalysis.pl in the topmost directories where you have unpacked the
# .rfa files you want analyzed. Start a command prompt window and go to this
# directory. Then type:
#
#     perl conanalysis.pl > bf1942con.txt
#
# and the bf1942con.txt will contain the analysis (which takes a bit to run).
#
# You can also run the program using just one or more directories, e.g.
#
#     perl conanalysis.pl objects/HandWeapons > hwcon.txt
#
# would analyze just the handweapons' files.


use strict;
use File::Find;

my $wiki;
$wiki = 0;	# for wiki output of hyperlink names; 0 == normal text output

my @confiles;
my $cfnum = 0;
my @dirs;
@dirs = @ARGV or $dirs[0] = '.';	# default is current directory
find( \&READRECURSIVEDIR, @dirs, );

my $fname;
my $path;
my $input;

my %fileuse ;
my %fileloc ;
my %thisfile ;
my %multiuses ;
my %firstuse ;
my %runcount ;
my %runs ;
my %tokenlist ;
my %dirtokenlist ;
my %arglist ;
my %tokentrans ;

my $prevfile = '' ;

my $qnan;
my $qnas;
my $qnaa;
my $qssn;
my $qsss;
my $qssa;

my $each;
my $always;

&PROCESSFILES();


sub READRECURSIVEDIR
{
	if ( m/\.(con|inc)$/ ) {
		$confiles[$cfnum] = $File::Find::name;
		$cfnum++;
	}
	# printf "that's good\n";
}

sub PROCESSFILES
{
	# get file name without path
	my $i;
	my @fld;
	my $filesfound;
	my $token ;
	my $lctoken ;
	my $prevlctoken ;
	my $spaces ;
	my $accumstring = '';
	my $diraccumstring = '';
	my $prevclass ;
	my $lcclass ;
	my @subfld;
	my $kk ;
	my $multi;

	# Read the .con and .inc files
	for ( $i = 0 ; $i < $cfnum ; $i++ ) {
		#printf "for file $confiles[$i]\n";
		@fld = split('/',$confiles[$i]);	# split
		$fname = lc($fld[$#fld]);
		$path = substr($confiles[$i],1,length($confiles[$i])-length($fname));
		#printf "file is $fname\n";

		$input = $confiles[$i];
		&READCONFILE();
	}

	# Output showing arguments for each property
	printf "List of the properties, and what arguments are used\n";
	printf "===================================================\n\n";
	printf "Number before colon is position of the argument after the property name,\nnumbers after are the values found for argument\n\n";
	my $maxfld ;
	$maxfld = 0 ;
	$prevlctoken = '';
	$prevclass = '';
	my @accumstringlist ;
	foreach $kk ( sort (keys %arglist) ) {
		@fld = split(' ',$kk);
		$token = $fld[0];
		$lctoken = $token;
		if ( $lctoken =~ m/\./ ) {
			@subfld = split('\.',$lctoken);
			$lcclass = $subfld[0];
		} else {
			$lcclass = $lctoken ;
		}
			
		if ( $prevlctoken ne $lctoken ) {
			if ( $prevlctoken ne '' ) {
				$spaces = ' ' x (length($prevlctoken)) ;
				for ( $i = 1 ; $i <= $maxfld ; $i++ ) {
					if ( $i == 1 ) {
						printf " $tokentrans{$prevlctoken} $i:$accumstringlist[$i]\n" ;
					} else {
						printf " $spaces $i:$accumstringlist[$i]\n" ;
					}
				}
			}
			$prevlctoken = $lctoken ;
			# $prevtoken = $token ;
			for ( $i = 1 ; $i <= $maxfld ; $i++ ) {
				$accumstringlist[$i] = '';
			}
			$maxfld = 0 ;
		}
		# put a line break between new classes
		if ( $prevclass ne $lcclass ) {
			printf "\n" ;
			$prevclass = $lcclass ;
		}
		# limit strings to around 400 characters or so - more than this is kind of crazy.
		if ( !exists($accumstringlist[$fld[1]]) || length($accumstringlist[$fld[1]]) < 400 ) {
			$accumstringlist[$fld[1]] .= ' ' . $fld[2] ;
			if ( length($accumstringlist[$fld[1]]) >= 400 ) {
				$accumstringlist[$fld[1]] .= ' ...' ;
			}
		}
		if ( $maxfld < $fld[1] ) {
			$maxfld = $fld[1] ;
		}
	}
	$spaces = ' ' x (length($prevlctoken)) ;
	for ( $i = 1 ; $i <= $maxfld ; $i++ ) {
		if ( $i == 1 ) {
			printf " $tokentrans{$prevlctoken} $i:$accumstringlist[$i]\n" ;
		} else {
			printf " $spaces $i:$accumstringlist[$i]\n" ;
		}
	}

	# now output the concordance
	printf     "##############################################\n";
	printf     "##############################################\n";
	printf "\n\nList of all properties and where they are used\n";
	printf     "==============================================\n\n";
	$prevlctoken = '';
	$prevclass = '';
	foreach $kk ( sort (keys %tokenlist) ) {
		@fld = split(' ',$kk);
		$token = $fld[0];
		$lctoken = $token;
		if ( $lctoken =~ m/\./ ) {
			@subfld = split('\.',$lctoken);
			$lcclass = $subfld[0];
		} else {
			$lcclass = $lctoken ;
		}
			
		if ( $prevlctoken ne $lctoken ) {
			if ( $prevlctoken ne '' ) {
				# 30 is the number of places for the token name
				$spaces = ' ' x (32 - length($prevlctoken)) ;
				chop $dirtokenlist{$token};
				printf " $tokentrans{$prevlctoken}%s  $accumstring($dirtokenlist{$prevlctoken})\n", $spaces ;
			}
			$prevlctoken = $lctoken ;
			# $prevtoken = $token ;
			$accumstring = '';
		}
		# put a line break between new classes
		if ( $prevclass ne $lcclass ) {
			printf "\n" ;
			$prevclass = $lcclass ;
		}
		$accumstring .= $fld[1] . ' ' ;
	}
	$spaces = ' ' x (25 - length($prevlctoken)) ;
	chop $dirtokenlist{$prevlctoken};
	printf " $tokentrans{$prevlctoken}%s  $accumstring($dirtokenlist{$prevlctoken})\n", $spaces ;

	printf "\n##############################################\n";
	printf   "##############################################\n\n";
	# output the individual .con file types and what they contain
	foreach $kk ( sort (keys %firstuse) ) {
		@fld = split(' ',$kk);
		$fname = $fld[0];
		if ( $prevfile ne $fname ) {
			if ( $prevfile ne '' ) {
				&OUTPUTDATA();
			}

			# now start the next one
			$prevfile = $fname ;
			$qnan = $qnas = $qnaa = $qssn = $qsss = $qssa = '';
			$filesfound = $fileuse{$fname} ;
			printf "========================\n\nFile $fname ";
			if ( $filesfound == 1 ) {
				printf "(1 occurrence) with full path $fileloc{$fname}\n\n";
				$each = '';
				$always = 'Used';
			} else {
				printf "($filesfound occurrences)\n\n";
				$each = ' each';
				$always = 'Always used';
			}
		}
		if ( exists( $multiuses{ $kk } ) ) {
			$multi = $multiuses{ $kk } ;
		} else {
			$multi = 0 ;
		}
		my $curtoken;
		$curtoken = $tokentrans{$fld[$#fld]} ;
		if ( $firstuse{ $kk } == $filesfound ) {
			# used in all files at least once
			if ( $multi == 0 ) {
				$qnan .= "$curtoken\n";
			} elsif ( $multi == $filesfound ) {
				$qnaa .= "$curtoken\n";
			} else {
				$qnas .= "$curtoken\n";
			}
		} else {
			# used in some files at least once
			if ( $multi == 0 ) {
				$qssn .= "$curtoken\n";
			} elsif ( $multi == $filesfound ) {
				$qssa .= "$curtoken\n";
			} else {
				$qsss .= "$curtoken\n";
			}
		}
	}
	&OUTPUTDATA();
}

sub READCONFILE
{
	unless (open(DATAFILE,$input)) {
		printf "Can't open $input: $!\n";
		exit 1 ;
	}
	my @fld;
	my @rfld;
	my $rem;
	my $obtoken;
	my $i;
	$fileuse{ $fname }++ ;
	$fileloc{ $fname } = $input ;
	while (<DATAFILE>) {
		chop;       # strip record separator
		@fld = split(' ',$_);

		if ( $#fld > 0 ) {

			$obtoken = lc($fld[0]) ;
			if ( substr($obtoken ,0,3) =~ /rem/i ) {
				# ignore - comment
			} elsif ( substr($obtoken ,0,1) eq "*" ) {
				# ignore - comment
			} elsif ( substr($obtoken ,0,1) eq "-" ) {
				# ignore - comment
			} elsif ( $obtoken eq "beginrem" ) {
				$rem = 1 ;
			} elsif ( $obtoken eq "endrem" ) {
				$rem = 0 ;
			} elsif ( $obtoken =~ /run/i ) {
				$runcount{ "$fname" }++ ;
				$runs{ "$fname $fld[1]" }++;
			} else {
				if ( exists( $thisfile{ "$input $obtoken" } ) ) {
					if ( $thisfile{ "$input $obtoken" } == 1 ) {
						$multiuses{ "$fname $obtoken" }++ ;
					}
				} else {
					$firstuse{ "$fname $obtoken" }++ ;
				}
				$tokentrans{$obtoken} = $fld[0] ;

				# adjust tokentrans for wiki output name
				if ( $wiki ) {
					my $class;
					my $subclass;
					my $lcclass;
					my $lcsubclass;
					my @subfld;
					if ( $fld[0] =~ m/\./ ) {
						@subfld = split('\.',$fld[0]);
						$class = $subfld[0];
						$lcclass = lc($class);
						$subclass = $subfld[1];
						$lcsubclass = lc($subclass);
						my $firstletter;
						# make first letter lowercase for comparison
						$class = lcfirst($class);
						$subclass = lcfirst($subclass);
						if ( $class eq $lcclass ) {
							$class = ucfirst($class);
							$class = '[[' . $class . ']]' ;
						} else {
							$class = ucfirst($class);
						}
						if ( $subclass eq $lcsubclass ) {
							$subclass = ucfirst($subclass);
							$subclass = '[[' . $subclass . ']]' ;
						} else {
							$subclass = ucfirst($subclass);
						}
						$tokentrans{$obtoken} = $class . '.' . $subclass ;
					}
				}

				$tokenlist{ "$obtoken $fname" }++;
				# make set of subdirectories for object file found
				my $subpath = $fileloc{ $fname };
				$subpath =~ s/ /_/g ;
				if ( substr($subpath,0,1) eq '.' ) {
					$subpath = substr($subpath,1);
				}
				if ( substr($subpath,0,1) eq '/' ) {
					$subpath = substr($subpath,1);
				}
				if ( $subpath =~ /\// ) {
					$subpath = $`;
					my $nextpath = $' ;
					if ( $nextpath =~ /\// ) {
						$subpath .= '/' . $`;
					}
				}
				if ( exists( $dirtokenlist{"$obtoken"} ) ) {
					# see if path is already on list
					my $found = 0 ;
					@rfld = split(' ',$dirtokenlist{"$obtoken"});
					my $ii;
					for ( $ii = 0 ; ($ii <= $#rfld) && ($found == 0); $ii++ ) {
						if ( lc($rfld[$ii]) eq lc($subpath) ) {
							$found = 1 ;
						}
					}
					if ( $found == 0 ) {
						$dirtokenlist{ "$obtoken" } .= $subpath . ' ';
					}
				} else {
					$dirtokenlist{"$obtoken"} = $subpath . ' ';
				}
				for ( $i = 1 ; $i <= $#fld ; $i++ ) {
					if ( $fld[$i] =~ m/\// ) {
						# x/y/z or x/y or x/y/z/w
						if ( index($fld[$i],'/') == rindex($fld[$i],'/') ) {
							$arglist{ "$obtoken $i x/y" }++;
						} else {
							my $subs = substr($fld[$i],index($fld[$i],'/')+1) ;
							if ( index($subs,'/') == rindex($subs,'/') ) {
								$arglist{ "$obtoken $i x/y/z" }++;
							} else {
								$arglist{ "$obtoken $i x/y/z/w" }++;
							}
						}
					} else {
						$arglist{ "$obtoken $i $fld[$i]" }++;
					}
				}
				$thisfile{ "$input $fld[0]" }++ ;
			}
		}
	}
}

sub OUTPUTDATA
{
	if ( $qnan ne '' || $qnas ne '' | $qnaa ne '' ) {
		printf "Required tokens for file\n------------------------\n\n";
		if ( $qnan ne '' ) {
			printf "Used once and only once in$each file:\n$qnan\n";
		}
		if ( $qnas ne '' ) {
			printf "Used once or more in$each file:\n$qnas\n";
		}
		if ( $qnaa ne '' ) {
			printf "$always multiple times in$each file:\n$qnaa\n";
		}
	}
	if ( $qssn ne '' || $qsss ne '' || $qssa ne '' ) {
		printf "Optional tokens for file\n------------------------\n\n";
		if ( $qssn ne '' ) {
			printf "Sometimes used once in$each file:\n$qssn\n";
		}
		if ( $qsss ne '' ) {
			printf "Sometimes used once or more in$each file:\n$qsss\n";
		}
		if ( $qssa ne '' ) {
			printf "Sometimes used multiple times in$each file (never just once):\n$qssa\n";
		}
	}
	my $jj;
	my @rfld;
	if ( exists( $runcount{ $prevfile } ) ) {
		my $str;
		$str = '';
		foreach $jj ( sort (keys %runs) ) {
			@rfld = split(' ',$jj);
			if ( $rfld[0] eq $prevfile ) {
				$str .= $rfld[1] . ' ' ;
			}
		}
		printf "\nRuns the files: $str\n";
	}
	printf "\n\n";
}