#!/usr/bin/perl -w

# Giant Grid O' Death program
# by Eric Haines, erich@acm.org
# September 28, 2003

# This Perl program analyzes a mod's .con and .inc files and produces a
# damage table and other related lists.
#
# You'll need Perl to run the damage.pl program. Just get it for free
# from Activestate, at
# http://www.activestate.com/Products/Download/Register.plex?id=ActivePerl
#
# To run, put damage.pl in the topmost directories where you have unpacked the
# .rfa files you want analyzed. Start a command prompt window and go to this
# directory. Then type:
#
#     perl damage.pl > damage.htm
#
# and the damage.htm will contain the analysis (which takes a bit to run).

use strict;
use File::Find;

# user options
my $web = 1;			# set to 1 for web output of table, 0 for text output

# Mod info
# set *AT MOST* one of these to 1:
my $bf = 1;		# set to 1 to present information relevant to BF 1942
my $dc = 0;		# set to 1 to present information relevant to Desert Combat
my $eod = 0;		# set to 1 to present information relevant to Eve of Destruction

my $findweaponnames = 1;	# set to 1 to find weapon names, 0 to never search but use defaults
my $forcearmornames = 0;	# set to 1 to force finding new armor names as available

my $minarmor = 40;	# minimum armor material to output
my $maxarmor = 66;	# maximum armor material to output
my $showallarmor = 0;	# set to 1 to show all armor types in range, even if not named

#-----------------------------------------------------------------------------
# end of user options
# Mess with the following code only if you know what the heck you're doing.

my @confiles;
my $cfnum = 0;
my @dirs ;

my $weapontype;
my $weaponname;

my $input;
my %attackname;
my %defendname;
my %materialdamage;
my %weaponatt;
my %damagemod;
my %moreweaponnames;
my %attacknumber;
my %attacknumber2;

my $validweapon;

my $showarmorchar = ($showallarmor ? "(unknown)" : "");

my %hp;

$hp{40} = $hp{41} = $hp{42} = 30;
$hp{43} = $hp{44} = $hp{45} = $hp{46} = $hp{47} = $hp{48} = $hp{49} = 50 ;
$hp{50} = $hp{51} = $hp{52} = $hp{53} = $hp{54} = 100 ;
$hp{55} = 200;
$hp{56} = 500 ;
$hp{57} = 200 ;
$hp{58} = 300 ;
$hp{59} = 500 ;
$hp{60} = $hp{61} = $hp{62} = 100 ;
$hp{63} = $hp{64} = $hp{65} = 130 ;
$hp{66} = 600 ;

my %deflabels;
my %attlabels;
my $minattgroup = 67;
my $maxattgroup;
my %hitpointsbynum;
my %suspect;

$deflabels{40} = "head";
$deflabels{41} = "trunk";
$deflabels{42} = "limbs";

$deflabels{43} = "kubel";
$deflabels{44} = $showarmorchar;
$deflabels{45} = "jeep [SPLASH]";
$deflabels{46} = "arty/APC";
$deflabels{47} = "arty/APC";
$deflabels{48} = $showarmorchar;
$deflabels{49} = $showarmorchar;
$deflabels{50} = "tank rear [SPLASH]";
$deflabels{51} = "tank side";
$deflabels{52} = "tank front";
$deflabels{53} = "tiger side";
$deflabels{54} = "tiger front";

if ( $dc ) {
$hp{45} = 100 ;
$hp{46} = $hp{47} = 100 ;
$deflabels{43} = $showarmorchar;
$deflabels{44} = $showarmorchar;
$deflabels{45} = "Humvee [SPLASH]";
$deflabels{46} = "APC armor";	# need a better term - used?
$deflabels{47} = "APC armor";	# need a better term - used?
$deflabels{48} = $showarmorchar;
$deflabels{49} = $showarmorchar;
$deflabels{50} = "tank rear [SPLASH]";
$deflabels{51} = "Shilka/MLRS [SPLASH]";
$deflabels{52} = "M2A3 [SPLASH]";
$deflabels{53} = "tank side";
$deflabels{54} = "T72/M1A1 [SPLASH]";
}
elsif ( $eod ) {
$hp{46} = $hp{47} = 100 ;
$deflabels{43} = $showarmorchar;
$deflabels{44} = $showarmorchar;
$deflabels{45} = "unarmored vehicle";
$deflabels{46} = "1-8mm";
$deflabels{47} = "9-20mm";
$deflabels{48} = $showarmorchar;
$deflabels{49} = $showarmorchar;
$deflabels{50} = "21-30mm";
$deflabels{51} = "31-60mm";
$deflabels{52} = "61-100mm";
$deflabels{53} = "101-150mm";
$deflabels{54} = "151-200mm";
}

# these tend to get used the same by many mods; set $forcearmornames to 1 if you don't like them
$deflabels{55} = "sub [ship SPLASH]";
$deflabels{56} = "under water";
$deflabels{57} = "turret [sub SPLASH]";
$deflabels{58} = "destroyer";
$deflabels{59} = "battleship";
$deflabels{60} = "engines [SPLASH]";
$deflabels{61} = "wings";
$deflabels{62} = $showarmorchar;
$deflabels{63} = "tail";
$deflabels{64} = $showarmorchar;
$deflabels{65} = $showarmorchar;
$deflabels{66} = "wood deck";


if ( $findweaponnames == 0 ) {

$attlabels{121} = "M3Grant Cannon";
$attlabels{122} = "M3Grant Gun";
$attlabels{123} = "M11-39 Cannon";
$attlabels{124} = "M11-39 Gun";
$attlabels{125} = "M3gmc";
$attlabels{126} = "Sturmgeschutz";
$attlabels{127} = "Mosquito MG";
$attlabels{128} = "BF110 MG";
$attlabels{129} = "Antitank Gun";
$attlabels{130} = "M3Grant splash";
$attlabels{131} = "M11-39 splash";
$attlabels{132} = "M3gmc splash";
$attlabels{133} = "Sturm splash";
$attlabels{134} = "A-T Gun splash";
$attlabels{199} = "AA splash";
$attlabels{200} = "Bazooka/Calliope splash";
$attlabels{201} = "Arty splash";
$attlabels{202} = "Light bomb";
$attlabels{203} = "";
$attlabels{204} = "Hvy bomb, ExpPack/Rocket splash";
$attlabels{205} = "Grenade";
$attlabels{206} = "Tank splash";
$attlabels{207} = "Hvy Tank splash";
$attlabels{208} = "Defgun/destroyer splash";
$attlabels{209} = "Battleship splash";
$attlabels{210} = "Knife";
$attlabels{211} = "";
$attlabels{212} = "Commando/Elite Knife Stab";
$attlabels{213} = "Plane MG";
$attlabels{214} = "Colt pistol";
$attlabels{215} = "P38 pistol";
$attlabels{216} = "SMG";
$attlabels{217} = "MP40 medic";
$attlabels{218} = "No.4 sniper";
$attlabels{219} = "K98 sniper";
$attlabels{220} = "";
$attlabels{221} = "";
$attlabels{222} = "BAR 1918";
$attlabels{223} = "StG 44";
$attlabels{224} = "Browning/SBD/B17";
$attlabels{225} = "MG42 machinegun";
$attlabels{226} = "Bazooka/Calliope";
$attlabels{227} = "";
$attlabels{228} = "AA gun";
$attlabels{229} = "";
$attlabels{230} = "(Landmine)";
$attlabels{231} = "";
$attlabels{232} = "Landmine splash";
$attlabels{233} = "";
$attlabels{234} = "";
$attlabels{235} = "";
$attlabels{236} = "Allied tank";
$attlabels{237} = "PanzerIV";
$attlabels{238} = "T35-85/M10";
$attlabels{239} = "Tiger";
$attlabels{240} = "(Hvy Bomb)";
$attlabels{241} = "(Wespe)";
$attlabels{242} = "(Lt. Bomb)";
$attlabels{243} = "";
$attlabels{244} = "";
$attlabels{245} = "";
$attlabels{246} = "";
$attlabels{247} = "";
$attlabels{248} = "Depth Charge splash";
$attlabels{249} = "";
$attlabels{250} = "Torpedo";
$attlabels{251} = "";
$attlabels{252} = "Destroyer/Defgun Projectile";
$attlabels{253} = "";
$attlabels{254} = "Battleship Projectile";
$attlabels{255} = "";
$attlabels{256} = "DP 1928/WasserFall";
$attlabels{257} = "Throwing Knife";
$attlabels{259} = "Shotgun";

$maxattgroup = 259 ;
}


# so it begins...

&PROCESSFILES();

&OUTPUTDATA();


sub PROCESSFILES
{
	# read game\bf1942\Game\materialManagerdefine.con for the MaterialDamage values.
	$input = "bf1942/Game/materialManagerdefine.con" ;
	&READMATERIALMANAGERDEFINE();

	undef(@dirs);
	undef( @confiles ) ;
	$cfnum = 0 ;
	$dirs[0] = "bf1942/Game/damage_system";
	find( \&READWEAPONSDIR, @dirs, );

	# read weapon's attack materials
	my $i;
	for ( $i = 0 ; $i < $cfnum ; $i++ ) {
#printf "for file $confiles[$i]\n";
		$input = $confiles[$i];
		&READMATERIALMANAGERDEFINE();
	}

	undef(@dirs);
	undef( @confiles ) ;
	$cfnum = 0 ;
	$dirs[0] = "objects";
	find( \&READWEAPONSDIR, @dirs, );

	# read weapon's attack materials
	for ( $i = 0 ; $i < $cfnum ; $i++ ) {
#printf "for file $confiles[$i]\n";
		$input = $confiles[$i];
		&READCONFILE();
	}

	undef(@dirs);
	undef( @confiles ) ;
	$cfnum = 0 ;
	$dirs[0] = "bf1942/Game/damage_system";
	find( \&READWEAPONSDIR, @dirs, );

	# read attack/defense matrix
	for ( $i = 0 ; $i < $cfnum ; $i++ ) {
#printf "for read matrix file $confiles[$i]\n";
		$input = $confiles[$i];
		&READMATRIXFILE();
	}
}

sub READWEAPONSDIR
{
	if ( m/\.(con|inc|rcm)$/ ) {
		$confiles[$cfnum] = $File::Find::name;
		$cfnum++;
	}
	# printf "that's good\n";
}


sub READCONFILE
{
	unless (open(DATAFILE,$input)) {
		printf "Can't open $input: $!\n";
		exit 1 ;
	}
	my @fld;
	$weapontype = $weaponname = 'none';
	$validweapon = 0 ;
	while (<DATAFILE>) {
		chop;       # strip record separator
		@fld = split(' ',$_);

		if ( $#fld > 0 ) {

			if ( lc($fld[0]) eq 'objecttemplate.create' ) {
#printf "creating $fld[1] $fld[2]\n";
				$weapontype = $fld[1] ;
				$weaponname = $fld[2] ;
				&CHECKWEAPONTYPE();
			}
			elsif ( $validweapon && lc($fld[0]) eq 'objecttemplate.material2' ) {
#printf "material2 $fld[1]\n";
				$attackname{$fld[1]} .= $weaponname . '/splash, ' ;
				$attacknumber2{lc($weaponname)} = $fld[1] ;
			}
			elsif ( $validweapon && lc($fld[0]) eq 'objecttemplate.material' ) {
#printf "material $fld[1]\n";
				# flaw in data files - filter it out
				if ( !($fld[1] == 216 && $weaponname eq 'sg44Projectile') ) {
					$attackname{$fld[1]} .= $weaponname . ', ' ;
					$attacknumber{lc($weaponname)} = $fld[1] ;
				}
			}
			elsif ( lc($fld[0]) eq 'objecttemplate.material' ) {
#printf "material $fld[1]\n";
				$defendname{$fld[1]} .= $weaponname . ', ' ;
				if ( $forcearmornames || ((!exists($deflabels{$fld[1]}) || $deflabels{$fld[1]} eq "") && $findweaponnames == 1 ) ) {
					$deflabels{$fld[1]} = $weaponname . " [SPLASH]" ;
				}
			}
			elsif ( lc($fld[0]) eq 'objecttemplate.projectiletemplate' ) {
				$moreweaponnames{$weaponname} .= $fld[1] ;
			}
			elsif ( lc($fld[0]) eq 'objecttemplate.maxhitpoints' ) {
				# $hitpoints{$weaponname} .= $fld[1] ;
				$hitpointsbynum{$fld[1]} .= $weaponname . ', ' ;
			}
		}
	}
	close(DATAFILE);
}

sub READMATERIALMANAGERDEFINE
{
	unless (open(DATAFILE,$input)) {
		printf "Can't open $input: $!\n";
		exit 1 ;
	}
	my @fld;
	my $materialid;
	$materialid = -1 ;
	while (<DATAFILE>) {
		chop;       # strip record separator
		@fld = split(' ',$_);

		if ( $#fld > 0 ) {

			if ( lc($fld[0]) eq 'materialmanager.materialattgroup' ) {
#printf("found material $fld[1]\n");
				$materialid = $fld[1] ;
			}
			# test for material2 first!
			elsif ( $materialid >= 0 && lc($fld[0]) eq 'materialmanager.materialdamage' ) {
				if ( exists( $materialdamage{$materialid} ) && $materialdamage{$materialid} != $fld[1] ) {
					printf STDERR "Warning: there are two or more definitions of the\n";
					printf STDERR "  MaterialDamage property for material $materialid.\n";
					printf STDERR "  MaterialDamage $materialdamage{$materialid} != $fld[1].\n";
					if ( $materialid == 256 ) {
						printf STDERR "  (BF 1942 has made this error with material 256,\n";
						printf STDERR "  which defines the DP 1928 and the Wasserfall.)\n";

					}
					elsif ( $materialid == 257 ) {
						printf STDERR "  (BF 1942 has made this error with material 257,\n";
						printf STDERR "  which defines the (unused) Bayonet and the Throwing Knife.)\n";

					} else {
						printf STDERR "  The results for material $materialid should be considered highly suspect!\n";
						$suspect{$materialid} = 1 ;
					}
				}
				$materialdamage{$materialid} = $fld[1] ;
			}
		}
	}
	close(DATAFILE);
}
sub READMATRIXFILE
{
	unless (open(DATAFILE,$input)) {
		printf "Can't open $input: $!\n";
		exit 1 ;
	}
	my @fld;
	my $fname;
	my $attgroup;
	my $defgroup;
	@fld = split('/',$input);	# split
	$fname = lc($fld[$#fld]);
	$fname = substr($fname,1,length($fname)-4);
	while (<DATAFILE>) {
		chop;       # strip record separator
		@fld = split(' ',$_);

		if ( $#fld > 0 ) {

			if ( lc($fld[0]) eq 'materialmanager.attgroup' ) {
				$attgroup = $fld[1] ;
				$weaponatt{$attgroup} = $fname ;
			}
			elsif ( lc($fld[0]) eq 'materialmanager.defgroup' ) {
				$defgroup = $fld[1] ;
			}
			elsif ( lc($fld[0]) eq 'materialmanager.damagemod' ) {
				$damagemod{$attgroup .' '. $defgroup} = $fld[1] ;
			}
			elsif ( lc($fld[0]) eq 'materialmanager.setcell' ) {
				$defgroup = $fld[1] ;
				$damagemod{$attgroup .' '. $defgroup} = $fld[2] ;
			}
		}
	}
	close(DATAFILE);
}

sub CHECKWEAPONTYPE
{
	$validweapon = 0 ;
	if ( $weapontype eq "Projectile" ) {
#	     $weapontype eq "HandFireArms" ) {
	    $validweapon = 1 ;
#printf "weapon is VALID\n";
	}
}
		

sub OUTPUTDATA
{
	my $kk ;
	my @fld;
	my $attgroup;
	my $defgroup;
	my $i;
	my %outtab;
	my $n;
	my %attused;

	# if we don't use the preset "prettified" names for the weapons, get our own at this point
	if ( $findweaponnames == 1 ) {
		# initialize any materials used to empty
		foreach $kk ( sort (keys %damagemod) ) {
			@fld = split(' ',$kk);	# split
			$attgroup = $fld[0];
			$attlabels{$attgroup} = "";
			$attused{$attgroup} = 1;
		}
		# now find a reasonable name
		foreach $kk ( sort numerically (keys %attackname) ) {
			if ( $kk >= $minattgroup ) {
				if ( $attackname{$kk} =~ /,/ ) {
					$attlabels{$kk} = $`;
				} else {
					$attlabels{$kk} = $attackname{$kk};
				}
# printf "for $attlabels{$kk} at $kk is $attackname{$kk}\n";
				$maxattgroup = $kk;
			}
		}
	}


	#printf ("\nMaterial Damage values\n");
	#printf (  "======================\n");
	#foreach $kk ( sort numerically (keys %materialdamage) ) {
	#	printf( " $kk: $materialdamage{$kk}\n");
	#}
	foreach $kk ( sort (keys %moreweaponnames) ) {
		if ( exists($attacknumber{lc($moreweaponnames{$kk})}) ) {
			$n = $attacknumber{lc($moreweaponnames{$kk})} ;
			$attackname{$n} .= '<i>'. $kk . '</i>, ';
		} else {
			#printf "error: number $n is for $kk and $moreweaponnames{$kk}\n";
		}
		if ( exists($attacknumber2{lc($moreweaponnames{$kk})}) ) {
			$n = $attacknumber2{lc($moreweaponnames{$kk})} ;
			$attackname{$n} .= '<i>'. $kk . '</i>, ';
		}
	}

	if ( $web ) {

if ( $dc ) {
printf "%s",<<"EOF";
<HTML><HEAD><TITLE>Giant Grid O' Death - Desert Combat</TITLE></HEAD>
<BODY>
<H1>Giant Grid O' Death - Desert Combat</H1>
EOF
} elsif ( $eod ) {
printf "%s",<<"EOF";
<HTML><HEAD><TITLE>Giant Grid O' Death - Eve of Destruction</TITLE></HEAD>
<BODY>
<H1>Giant Grid O' Death - Eve of Destruction</H1>
EOF
} else {
printf "%s",<<"EOF";
<HTML><HEAD><TITLE>Giant Grid O' Death</TITLE></HEAD>
<BODY>
<H1>Giant Grid O' Death</H1>
EOF
}

		if ( $bf ) {
printf "%s",<<"EOF";
<B>Secret Weapons attack effectiveness:</B> The Calliope tank projectiles are the same as bazooka rounds (which is why they stink against infantry). The Natter, Horton HO229 flying wing, and Armstrong Whitworth AW-52 aircraft have rockets that act like the heavy bombs in the B17, which are each equivalent to an exppack. The K98 rifle grenade has the same effect as a normal grenade. The T95 heavy allied tank has the same shell as a defgun and destroyer. The SturmTiger's projectile is the same as a battleship's "fat" projectile (material 209). The commando and elite knives kill in a stab to any part of the body. When thrown, they kill in one hit to anywhere above the waist. The Goblin's machineguns are like most every other airplane's machinegun. The Gewehr 42's ammo does the same damage that a Sg 44 does. Similarly, the Sten has the same ammo as other Medic's guns, the Bren LMG has the same ammo as the BAR (but its effectiveness starts to decrease at 50 yards, and drops to half at 100 yards), and the Gewehr 43 has the same as a K98. The flak panzer, you'll be shocked to find, shoots the same rounds as every other flak gun. The Wasserfall explosion kills infantry it hits, can damage jeeps and APCs but not tanks, and is most effective against aircraft engines; however, the "body" of the rocket itself appears to cause collision damage to tanks. The shotgun shoots a pentagonal pattern of 5 pellets and kills with 2 pellets to the head, or 4 to the trunk, or 6 to the legs; it is also effective against motorcycles, jeeps, and planes, but nothing else.
<P>
<B>Secret Weapons defense effectiveness:</B> Here's the new hitpoint rundown. In the air: C47 takes 250 HP; AW52 takes 160; Horton takes 60; Goblin takes 40; Natter takes 15 (!); Wasserfall Rocket takes 10. On the ground: SturmTiger takes 100 (like a regular Tiger); Flakpanzer takes 70; LVT-2 Allied Amphibious takes 65; Schwimmwagen jeep takes 50 (same as a regular jeep); the motorcycles take 40. Inflatable boats (at Peenem&uuml;nde and Telemark) take 50 HP and defend like the stronger APC armor (material 45). The Wasserfall launching station takes 50 hitpoints and defends with material 45 (like armor on an APC). The jetpack itself (when not in use) takes 100 and defends like a plane engine. Objectives: Eagle's Nest door and safe both take 500 HP (this translates to 7 exppacks each); Essen fuel silos take 600 HP (which translates to about 15 exppacks, since you can't place the exppacks close to the center of these); the Hellendoorn V2s take 600 HP each; the Kbely Airfield UFOs take 675 HP each; the Mimoyecques vents each take 550 HP; the Telemark Turbines take 675 HP each.
<P>
<B>Changes going from version 1.31 to 1.4:</B> DICE's release notes state that bayonet damage has increased, but it has not actually changed, due to a bug in hooking up the damage model.
The sniper rifle now reloads as fast as the engineer's rifle did in 1.31, i.e. it now reloads in 1.6 seconds instead of 3 seconds.
Machineguns are less potent against planes: hitting in the engine causes 12.5% less damage, hitting the wings causes 25% less damage, hitting the tail causes 10% less damage. The Tiger tank now has 125 hitpoints; it used to have 100 (note that this fact is not reflected in the tables that follow).
A minor inconsistency, that the K98 could damage jeeps and the No.4 could not, has been corrected; now both can. Flak's computed damage values have not changed, but whatever is inside its radius of 20 meters gets hit with full damage by the exploding shell; it used to be that damage fell off with distance. Grenades and exppacks no longer can launch people into the air.
There are new assault rifles: the DP 1928 Russian assault rifle has a little less zoom-in, but has exactly the same deviation as a BAR.
The Type 99 Japanese assault rifle is also like a BAR in many ways. However, like a medic's gun, its damage starts to drop off at 50 meters, going to a minimum of half damage above 100 meters. To compensate, it has somewhat less deviation when aiming and when moving. It also has a little less zoom-in than the BAR.
<P>
EOF

}	# end if $bf

printf "%s",<<"EOF";
What is listed here is the maximum damage for a weapon. Splash damage usually drops off with the distance from impact, the exceptions being for flak and landmines. Weapons such as the Breda, medics' guns, and pistols drop off with range. Also, the angle at which a shot hits an armored surface matters: dead on gives full damage, a glancing angle does less (it's the cosine of the angle compared to the surface normal). For weapons with splash, look at the defense material list at the very bottom of this page - this material is the one used as the defense when doing splash damage; ignore the location-specific material.
<P>
EOF

		printf ("<table BORDER=1 CELLSPACING=1 CELLPADDING=2 >\n");

		my $prevattgroup ;
		$prevattgroup = -999 ;

		printf "<tr>\n<td></td>\n";
		for ( $i = $minarmor ; $i <= $maxarmor ; $i++ ) {
			if ( $deflabels{$i} ne "" ) {
				printf "<td><B>$i</B><br>$deflabels{$i}</td>\n";
				$outtab{$i} = '<i>'.$hp{$i}.'</i>' ;
			}
		}
		printf "</tr>\n";

		# output the lines of the major damage table
		my $problem = '';
		foreach $kk ( sort (keys %damagemod) ) {
			@fld = split(' ',$kk);	# split
			$attgroup = $fld[0];
			$defgroup = $fld[1];
			if ( $attgroup >= $minattgroup && $attgroup <= $maxattgroup ) {
				# it's a weapon
 				if ( $prevattgroup != $attgroup ) {
 					printf "<tr>\n";
					if ( $prevattgroup == -999 ) {
 						printf "<td>defender's typical hitpoints:</td>\n";
					} else {
 						printf "<td><B>$prevattgroup</B><br>$attlabels{$prevattgroup}$problem</td>\n";
					}
					for ( $i = $minarmor ; $i <= $maxarmor ; $i++ ) {
						if ( $deflabels{$i} ne "" ) {
							printf " <td>$outtab{$i}</td>";
							$outtab{$i} = ""; # or $materialdamage{$attgroup};
						}
					}
					if ( $prevattgroup == -999 ) {
 						printf "<td>defender's hitpoints:</td>\n";
					} else {
 						printf "<td><B>$prevattgroup</B><br>$attlabels{$prevattgroup}$problem</td>\n";
					}
 					printf "</tr>\n";
 					$prevattgroup = $attgroup ;

					# set up problem variable for next line
					if ( exists( $suspect{$attgroup} ) ) {
						$problem = "<br><font color=\"#FF0000\">highly suspect info!</font>";
					} else {
						$problem = '';
					}
 				}
				if ( $defgroup >= $minarmor && $defgroup <= $maxarmor ) {
				    if ( $deflabels{$defgroup} ne "" ) {
					# it's something killable
					my $val;
					$val = $materialdamage{$attgroup} * $damagemod{$kk} ;
					my $hits;
					if ( $val != 0 ) {
						$hits = $hp{$defgroup} / $val;
						my $clr;
						$clr = "#353CCB";
						if ( $hits <= 1 ) {
							$clr = "#FF0000";
						} elsif ( $hits <= 2 ) {
							$clr = "#FF9409";
						} elsif ( $hits <= 3 ) {
							$clr = "#CB9F35";
						} elsif ( $hits <= 5 ) {
							$clr = "#89B64E";
						} elsif ( $hits <= 10 ) {
							$clr = "#35CBB1";
						}
						$outtab{$defgroup} = sprintf "$val (<font color=\"$clr\">%5.1f</font>)",$hits;
					} else {
						$outtab{$defgroup} = $val ;
					}
				    }
				}
			}			
		}
		# output last line (really should be a subroutine...)
 		printf "<tr>\n";
		printf "<td><B>$prevattgroup</B><br>$attlabels{$prevattgroup}$problem</td>\n";
		for ( $i = $minarmor ; $i <= $maxarmor ; $i++ ) {
			if ( $deflabels{$i} ne "" ) {
				printf " <td>$outtab{$i}</td>";
			}
		}
		printf "<td><B>$prevattgroup</B><br>$attlabels{$prevattgroup}$problem</td>\n";
		printf "</tr>\n";
		# output bottom defense armor line
		printf "<tr>\n<td></td>\n";
		for ( $i = $minarmor ; $i <= $maxarmor ; $i++ ) {
			if ( $deflabels{$i} ne "" ) {
				printf "<td><B>$i</B><br>$deflabels{$i}</td>\n";
			}
		}
		printf "</tr>\n";
		printf "</table>\n";
	} else {
		printf ("\nWeapon vs. Material\n");
		printf (  "===================\n");
		foreach $kk ( sort (keys %damagemod) ) {
			@fld = split(' ',$kk);	# split
			$attgroup = $fld[0];
			$defgroup = $fld[1];
			if ( $attgroup >= $minattgroup && $attgroup <= $maxattgroup ) {
				# it's a weapon
				if ( $defgroup >= $minarmor && $defgroup <= $maxarmor ) {
					# it's something killable
					printf ("Weapon $weaponatt{$attgroup} ID $attgroup attacks material $defgroup:\n    MaterialDamage is $materialdamage{$attgroup} and DamageMod is $damagemod{$kk} giving %f damage\n", $materialdamage{$attgroup} * $damagemod{$kk} );
				}
			}			
		}
	}

	if ( $web ) {
		printf "<H2>Weapon Material IDs</H2>\n";

		printf "These are derived from the various weapons files. The italicized names are the weapons that use the projectiles.\n";

		if ( $bf ) {
		printf "They're displayed so you can find things like the Breda, which uses BAR bullets. Some names should not be taken too seriously, like the bayonets are actually knives but are associated with guns. For bayonets, instead look for the names ending with <i>StabArm</i> (which are all Material 210).\n" ;
		}

		printf " Weapons names that appear in parentheses are weapons that are partially defined, but have no damage modifiers (and so cannot actually damage anything).\n<P>\n";

		foreach $kk ( sort numerically (keys %attackname) ) {
			if ( $kk > 100 ) {
				chop $attackname{$kk} ;
				chop $attackname{$kk} ;
				if ( exists( $attused{$kk} ) ) {
					printf( " $kk: $attackname{$kk}<br>\n");
				} else {
					printf( " ($kk: $attackname{$kk})<br>\n");
				}
			}
		}

if ( $bf ) {

printf "%s",<<"EOF";

<H2>Direct Hit Defense Material IDs</H2>
These are derived from opening up the collision meshes of objects and looking at the materials assigned to the various faces. This is an ongoing process searching these out, so don't trust all these values.
<P>
40: Soldier (Head)<br>
41: Soldier (Body)<br>
42: Soldier (Arms/Legs)<br>
43: Kubelwagen (Hood/Lower Front)<br>
44: (unused)<br>
45: Willy (All), Kubelwagen (most of Body), Landing Crafts (All)<br>
46: Halftracks (Sides/Rear), Katyusha (Body), Wespe (Turret), Priest (Back/Upper Sides)<br>
47: Halftracks (Hood), Priest (Top Front)<br>
48: (unused)<br>
49: (unused)<br>
50: Sherman (Rear), Chi-Ha (Top/Rear), Flak Gun (All), Panzer IV (Top/Rear), T34-85 (Top/Rear)<br>
51: Sherman (Side/Top/Tracks), Halftracks (Tracks/Wheels), Tiger (Rear/Tracks) Chi-Ha (Side), M10 (Top/Back/Tracks), Panzer IV (Sides/Tracks), T34 (Top/Rear/Tracks), T34-85 (Sides/Tracks), Wespe (Front/Tracks), Priest (Bottom Front/Tracks)<br>
52: Sherman (Front), Chi-Ha (Front), Panzer IV (Front), T34-85 (Front)<br>
53: Tiger (Top/Side), Defgun (All), M10 (Sides), T34 (Sides)<br>
54: Tiger (Front), M10 (Front), T34 (Front)<br>
55: Submarine (Hull)<br>
56: Ships (Underwater), Submarine (Conning Tower)<br>
57: Destroyer (Turret)<br>
58: Destroyer (Hull/Super)<br>
59: Battleship (Hull/Super), Carrier (Hull/Super)<br>
60: Mustang/Corsair/Spitfire/Yak9/Ju87/Zero/B109 (Engines/Cockpit), Zero (Fuselage), SBD (All)<br>
61: Most Planes (Wings), Mustang (Tail)<br>
62: Mustang (Wings), B17 (Tail)<br>
63: AichiVal/Stuka (Cockpit), Most Planes (Tails)<br>
64: (unused)<br>
65: (unused)<br>
66: Ships (Wooden Decks)<br>
<P>
This information is summarized in more precise (and lengthier) fashion in this table:
<P>
<table BORDER=1 CELLSPACING=1 CELLPADDING=1 > 
<tr>
  <td></td>
  <td ALIGN="CENTER" colspan=4><B><I>Unarmored</I></B></td>
  <td ALIGN="CENTER" colspan=5><B><I>Armor</I></B></td>
  <td ALIGN="CENTER" colspan=5><B><I>Ship</B></td>
  <td ALIGN="CENTER" colspan=3><B><I>Airplane</I></B></td>
  <td ALIGN="CENTER"><B><I>Wooden Deck</I></B></td>
</tr>
<tr>
  <td></td>
  <td ALIGN="CENTER"><B>43</B></td>
  <td ALIGN="CENTER"><B>45<br>(splash)</B></td>
  <td ALIGN="CENTER"><B>46</B></td>
  <td ALIGN="CENTER"><B>47</B></td>
  <td ALIGN="CENTER"><B>50<br>(splash)</B></td>
  <td ALIGN="CENTER"><B>51</B></td>
  <td ALIGN="CENTER"><B>52</B></td>
  <td ALIGN="CENTER"><B>53</B></td>
  <td ALIGN="CENTER"><B>54</B></td>
  <td ALIGN="CENTER"><B>55<br>(splash)</B></td>
  <td ALIGN="CENTER"><B>56</B></td>
  <td ALIGN="CENTER"><B>57<br>(splash)</B></td>
  <td ALIGN="CENTER"><B>58</B></td>
  <td ALIGN="CENTER"><B>59</B></td>
  <td ALIGN="CENTER"><B>60<br>(splash)</B></td>
  <td ALIGN="CENTER"><B>61</B></td>
  <td ALIGN="CENTER"><B>63</B></td>
  <td ALIGN="CENTER"><B>66</B></td>
</tr>
<tr>
  <td>Kubelwagen</td>
  <td ALIGN="CENTER">front hood, front upper sides</td>
  <td ALIGN="CENTER">all else</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
</tr>
<tr>
  <td>Willy Jeep</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER">all</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
</tr>
<tr>
  <td>Katyusha</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER">all</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
</tr>
<tr>
  <td>Priest</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER">all else</td>
  <td ALIGN="CENTER">front</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER">tracks</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
</tr>
<tr>
  <td>Wespe</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER">all else</td>
  <td ALIGN="CENTER">bottom front</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER">front top, tracks</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
</tr>
<tr>
  <td>APCs (except M3 GMC)</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER">all else</td>
  <td ALIGN="CENTER">front, hood</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER">tracks</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
</tr>
<tr>
  <td>M3 GMC</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER">lower front, all else</td>
  <td ALIGN="CENTER">upper front, hood</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER">tracks</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
</tr>
<tr>
  <td>Chi-ha</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER">rear, turret, hull top, bottom</td>
  <td ALIGN="CENTER">tracks, cannon</td>
  <td ALIGN="CENTER">front, turret sides</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
</tr>
<tr>
  <td>T34/85</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER">turret rear and top, all else</td>
  <td ALIGN="CENTER">front top, sides, cannon</td>
  <td ALIGN="CENTER">front, turret front and sides</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
</tr>
<tr>
  <td>Carro Armato</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER">rear, bottom</td>
  <td ALIGN="CENTER">sides, rear top</td>
  <td ALIGN="CENTER">front</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
</tr>
<tr>
  <td>M3 Grant</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER">rear</td>
  <td ALIGN="CENTER">all else</td>
  <td ALIGN="CENTER">front</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
</tr>
<tr>
  <td>Sherman</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER">all else</td>
  <td ALIGN="CENTER">sides</td>
  <td ALIGN="CENTER">front, turret sides</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
</tr>
<tr>
  <td>Sturmgeschutz</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER">rear, sides</td>
  <td ALIGN="CENTER">gunner</td>
  <td ALIGN="CENTER">all else</td>
  <td ALIGN="CENTER">front</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
</tr>
<tr>
  <td>T34</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER">turret rear and top, all else</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER">upper sides, front top</td>
  <td ALIGN="CENTER">front, turret front and sides</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
</tr>
<tr>
  <td>M10</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER">turret top, all else</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER">upper sides</td>
  <td ALIGN="CENTER">front, turret sides</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
</tr>
<tr>
  <td>Tiger</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER">turret top, all else</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER">upper sides, front top</td>
  <td ALIGN="CENTER">front, turret sides</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
</tr>
<tr>
  <td>AA flak gun</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER">all</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
</tr>
<tr>
  <td>Anti-Tank</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER">barrel, all else</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER">front armor</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
</tr>
<tr>
  <td>Defgun</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER">all</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
</tr>
<tr>
  <td>landing craft</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER">all else</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER">ramp</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
</tr>
<tr>
  <td>Sub VII</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER">hatch</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER">hull</td>
  <td ALIGN="CENTER">conning tower</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
</tr>
<tr>
  <td>Gato</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER">hull</td>
  <td ALIGN="CENTER">conning tower</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
</tr>
<tr>
  <td>Fletcher</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER">underwater</td>
  <td ALIGN="CENTER">turret</td>
  <td ALIGN="CENTER">all else</td>
  <td ALIGN="CENTER">cannon</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
</tr>
<tr>
  <td>Hatsuzuki</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER">underwater</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER">all else</td>
  <td ALIGN="CENTER">turret</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
</tr>
<tr>
  <td>Battleship, Carrier</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER">underwater</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER">sides, conning tower</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER">deck</td>
</tr>
<tr>
  <td>SBD</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER">all else</td>
  <td ALIGN="CENTER">wings, rear fuselage</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
</tr>
<tr>
  <td>Mosquito, BF110</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER">engines</td>
  <td ALIGN="CENTER">all else</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
</tr>
<tr>
  <td>AichiVal, B109, Corsair, Junkers 87, Yak 9</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER">engine, cockpit</td>
  <td ALIGN="CENTER">wings, rear</td>
  <td ALIGN="CENTER">tail</td>
  <td ALIGN="CENTER"></td>
</tr>
<tr>
  <td>Zero</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER">fuselage</td>
  <td ALIGN="CENTER">wings</td>
  <td ALIGN="CENTER">tail</td>
  <td ALIGN="CENTER"></td>
</tr>
<tr>
  <td>Mustang</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER">engine</td>
  <td ALIGN="CENTER">all else</td>
  <td ALIGN="CENTER">tail</td>
  <td ALIGN="CENTER"></td>
</tr>
<tr>
  <td>B17</td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER"></td>
  <td ALIGN="CENTER">engines</td>
  <td ALIGN="CENTER">wings, sides</td>
  <td ALIGN="CENTER">all else</td>
  <td ALIGN="CENTER"></td>
</tr>
<tr>
  <td></td>
  <td ALIGN="CENTER"><B>43</B></td>
  <td ALIGN="CENTER"><B>45<br>(splash)</B></td>
  <td ALIGN="CENTER"><B>46</B></td>
  <td ALIGN="CENTER"><B>47</B></td>
  <td ALIGN="CENTER"><B>50<br>(splash)</B></td>
  <td ALIGN="CENTER"><B>51</B></td>
  <td ALIGN="CENTER"><B>52</B></td>
  <td ALIGN="CENTER"><B>53</B></td>
  <td ALIGN="CENTER"><B>54</B></td>
  <td ALIGN="CENTER"><B>55<br>(splash)</B></td>
  <td ALIGN="CENTER"><B>56</B></td>
  <td ALIGN="CENTER"><B>57<br>(splash)</B></td>
  <td ALIGN="CENTER"><B>58</B></td>
  <td ALIGN="CENTER"><B>59</B></td>
  <td ALIGN="CENTER"><B>60<br>(splash)</B></td>
  <td ALIGN="CENTER"><B>61</B></td>
  <td ALIGN="CENTER"><B>63</B></td>
  <td ALIGN="CENTER"><B>66</B></td>
</tr>
<tr>
  <td></td>
  <td ALIGN="CENTER" colspan=4><B><I>Unarmored</I></B></td>
  <td ALIGN="CENTER" colspan=5><B><I>Armor</I></B></td>
  <td ALIGN="CENTER" colspan=5><B><I>Ship</B></td>
  <td ALIGN="CENTER" colspan=3><B><I>Airplane</I></B></td>
  <td ALIGN="CENTER"><B><I>Wooden Deck</I></B></td>
</tr>
</table>
<P>
Note: peculiarly, the Hatsuzuki has a material of 86 for its guns; and it basically just makes these something that cannot be hit (though the turret can be hit).
EOF
}

		printf "<H2>Vehicle Splash Defense Material IDs</H2>\n";
		printf "These are set by ObjectTemplate.Material in the various vehicle Objects.con files, but are not for direct (non-splash) damage. These materials are used for computing splash damage effects.\n<P>\n";
		foreach $kk ( sort numerically (keys %defendname) ) {
			# +3 is to get past soldier body
			if ( $kk >= $minarmor+3 && $kk <= $maxarmor ) {
				my $trimmed = $defendname{$kk};
				chop $trimmed;
				chop $trimmed;
				printf( " $kk: $trimmed<br>\n");
			}
		}

	} else {
		printf ("\nWeapon Material IDs\n");
		printf (  "===================\n");
		foreach $kk ( sort numerically (keys %attackname) ) {
			my $cleanup = $attackname{$kk} ;
			$cleanup =~ s/<i>//g;
			$cleanup =~ s/<\/i>//g;
			printf( " $kk: $cleanup\n");
		}
	}

	if ( $web ) {
		printf "<H2>Hitpoints</H2>\n";
		printf "This section lists the number of hitpoints the various objects take.\nNote that this list is auto-generated, so there may be some entries that are not meaningful within the current game. For example, there is a window hitpoint value, implying that at one time DICE considered having destructible windows.<p>\n";
		foreach $kk ( sort numerically (keys %hitpointsbynum) ) {
			my $val;
			$val = $hitpointsbynum{$kk};
			chop $val;
			chop $val;
			printf( " $kk: $val<br>\n");
		}
	} else {
		printf ("\nHitpoints\n");
		printf (  "=========\n");
		foreach $kk ( sort numerically (keys %hitpointsbynum) ) {
			my $val;
			$val = $hitpointsbynum{$kk};
			chop $val;
			chop $val;
			printf( "$kk: $val\n");
		}
	}

	if ( $web ) {
printf( "<HR>\n<ADDRESS>\nEric Haines / <A HREF=\"mailto:erich\@acm.org\">erich\@acm.org</A>\n</ADDRESS>\n</BODY>\n</HTML>\n" ) ;
	}
}

sub numerically { $a <=> $b ; }
